﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEditor;

namespace BS
{
    public class NodalPathfinding2D : MonoBehaviour
    {
        public Bounds bounds;
        [Range(0.1f, 1f)]
        public float pointInterval = 0.5f;

        
        public LayerMask blockMask;
        private Dictionary<Vector2Int, Node> nodes = new Dictionary<Vector2Int, Node>();
        public bool isBaked { get { return nodes.Count > 0; } }

        public bool enableGizmos = true;


        private void OnDrawGizmos()
        {
            if (!enableGizmos)
                return;
            if (isBaked)
            {
                int xCount = Mathf.FloorToInt((bounds.max.x - bounds.min.x) / pointInterval) + 1;
                int yCount = Mathf.FloorToInt((bounds.max.y - bounds.min.y) / pointInterval) + 1;
                Gizmos.color = Color.green;
                foreach (var node in nodes.Values)
                {
                    Gizmos.color = new Color(node.gridPosition.x / (float)xCount, node.gridPosition.y / (float)yCount, 0, 1);
                    Gizmos.DrawSphere(node.worldPositon, pointInterval / 5);
                    foreach (var adj in node.adjacencies)
                    {
                        Gizmos.DrawLine(node.worldPositon, adj.worldPositon);
                    }
                }
            }
            else
            {
                Gizmos.DrawWireCube(bounds.center, bounds.size);

                for (float i = bounds.min.x; i <= bounds.max.x; i += pointInterval)
                {
                    for (float j = bounds.min.y; j <= bounds.max.y; j += pointInterval)
                    {
                        if (Physics2D.OverlapPoint(new Vector2(i, j), blockMask))
                            Gizmos.color = Color.red;
                        else
                            Gizmos.color = Color.green;
                        Gizmos.DrawSphere(new Vector3(i, j, 0), pointInterval / 5);
                    }
                }

                for (float i = bounds.min.x + pointInterval / 2; i <= bounds.max.x; i += pointInterval)
                {
                    for (float j = bounds.min.y + pointInterval / 2; j <= bounds.max.y; j += pointInterval)
                    {
                        if (Physics2D.OverlapPoint(new Vector2(i, j), blockMask))
                            Gizmos.color = Color.red;
                        else
                            Gizmos.color = Color.green;
                        Gizmos.DrawSphere(new Vector3(i, j, 0), pointInterval / 5);
                    }
                }
            }
        }

        public void BakeNodes()
        {
            Debug.Log("Baking Start");
            ClearNodes();
            int xCount = Mathf.FloorToInt((bounds.max.x - bounds.min.x) / pointInterval) + 1;
            int yCount = 2 * Mathf.FloorToInt((bounds.max.y - bounds.min.y) / pointInterval) + 1;

            for (int i = 0; i < xCount; i++)
            {
                for (int j = 0; j < yCount; j++)
                {
                    if (j % 2 == 1 && i + 1 == xCount)
                        continue;
                    Vector3 worldPosition = bounds.min + new Vector3(i * pointInterval + (j % 2 == 0 ? 0 : pointInterval / 2), j * pointInterval / 2);
                    if (Physics2D.OverlapPoint(worldPosition, blockMask))
                        continue;
                    Node newNode = new Node(new Vector2Int(i, j), worldPosition);
                    nodes.Add(newNode.gridPosition, newNode);
                }
                
            }

            for (int i = 0; i < xCount; i++)
            {
                for (int j = 0; j < yCount; j++)
                {
                    List<Node> adj = new List<Node>();
                    Vector2Int pos = new Vector2Int(i, j);
                    if (!nodes.ContainsKey(pos))
                        continue;
                    Node node = nodes[pos];
                    if (nodes.ContainsKey(node.gridPosition + new Vector2Int(1, 0))) //Right
                    {
                        adj.Add(nodes[node.gridPosition + new Vector2Int(1, 0)]);
                    }
                    if (nodes.ContainsKey(node.gridPosition + new Vector2Int(-1, 0))) //Left
                    {
                        adj.Add(nodes[node.gridPosition + new Vector2Int(-1, 0)]);
                    }
                    if (nodes.ContainsKey(node.gridPosition + new Vector2Int(0, 2))) // Up
                    {
                        adj.Add(nodes[node.gridPosition + new Vector2Int(0, 2)]);
                    }
                    if (nodes.ContainsKey(node.gridPosition + new Vector2Int(0, -2))) // Down
                    {
                        adj.Add(nodes[node.gridPosition + new Vector2Int(0, -2)]);
                    }

                    if (nodes.ContainsKey(node.gridPosition + new Vector2Int(-1 + node.gridPosition.y % 2, 1))) //UpLeft
                    {
                        adj.Add(nodes[node.gridPosition + new Vector2Int(-1 + node.gridPosition.y % 2, 1)]);
                    }
                    if (nodes.ContainsKey(node.gridPosition + new Vector2Int(node.gridPosition.y % 2, 1))) //UpRight
                    {
                        adj.Add(nodes[node.gridPosition + new Vector2Int(node.gridPosition.y % 2, 1)]);
                    }
                    if (nodes.ContainsKey(node.gridPosition + new Vector2Int(-1 + node.gridPosition.y % 2, -1))) //DownLeft
                    {
                        adj.Add(nodes[node.gridPosition + new Vector2Int(-1 + node.gridPosition.y % 2, -1)]);
                    }
                    if (nodes.ContainsKey(node.gridPosition + new Vector2Int(node.gridPosition.y % 2, -1))) //DownRight
                    {
                        adj.Add(nodes[node.gridPosition + new Vector2Int(node.gridPosition.y % 2, -1)]);
                    }
                    node.adjacencies = adj;
                    
                    nodes[pos] = node;
                }
            }
            Debug.Log("Baking End");
        }

        public void ClearNodes()
        {
            nodes.Clear();
        }
    }

    struct Node
    {
        public Node(Vector2Int gridPosition, Vector3 worldPositon)
        {
            this.gridPosition = gridPosition;
            this.worldPositon = worldPositon;
            adjacencies = null;
        }
        public Vector2Int gridPosition;
        public Vector3 worldPositon;
        public List<Node> adjacencies;
    }
}